<?php

namespace Drupal\cl_selector_field\Plugin\Field\FieldType;

use Drupal\cl_components\Component\Component;
use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\ComponentNotFoundException;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemBase;
use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\TypedData\DataDefinition;

/**
 * Defines the 'cl_selector_field_style_selector' field type.
 *
 * @FieldType(
 *   id = "cl_selector_field_style_selector",
 *   label = @Translation("Style Selector"),
 *   category = @Translation("General"),
 *   default_widget = "cl_selector_field_style_selector",
 *   default_formatter = "cl_selector_field_style_selector_table"
 * )
 */
class StyleSelectorItem extends FieldItemBase {

  const VARIANT = 'variant';

  /**
   * {@inheritdoc}
   */
  public static function defaultFieldSettings() {
    $settings = [
      'features' => [
        static::VARIANT => static::VARIANT,
      ],
      'forbidden' => [],
      'allowed' => [],
    ];
    return $settings + parent::defaultFieldSettings();
  }

  /**
   * {@inheritdoc}
   */
  public function fieldSettingsForm(array $form, FormStateInterface $form_state) {
    $settings = $this->getSettings();

    $element['features'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Supported features'),
      '#description' => $this->t('The component is mandatory, but the variant support is optional for each field instance.'),
      '#default_value' => $settings['features'],
      '#options' => [
        static::VARIANT => $this->t('Variant'),
      ],
    ];
    $options = static::allAllowedComponentValues();
    $element['forbidden'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Forbidden components'),
      '#description' => $this->t('When showing the component dropdown, all components will be displayed except the components selected here. This is used to limit the components available for a given content type.'),
      '#options' => $options,
      '#default_value' => $settings['forbidden'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $element['allowed'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Allowed components'),
      '#description' => $this->t('When showing the component dropdown, only the components selected here will be displayed as an option. <strong>NOTE:</strong> "forbidden" components will not be shown even when allowed here.'),
      '#options' => $options,
      '#default_value' => $settings['allowed'],
      '#attributes' => ['class' => ['long-checkboxes-list']],
    ];
    $element['#attached']['library'][] = 'cl_selector_field/cl_selector_field_style_selector';

    return $element;
  }

  /**
   * Returns all the allowed values for 'component' sub-field.
   *
   * @param array $allowed
   *   The allowed components.
   * @param array $forbidden
   *   The forbidden components.
   *
   * @return array
   *   The list of allowed values.
   */
  public static function allAllowedComponentValues(array $allowed = [], array $forbidden = []): array {
    $options = static::allAllowedComponentValuesSelect($allowed, $forbidden);
    return array_reduce(
      $options,
      static function (array $options, array $group) {
        return array_merge($options, $group);
      },
      []
    );
  }

  /**
   * Returns all the allowed values for 'component' sub-field grouped by letter.
   *
   * @param array $allowed
   *   The allowed components.
   * @param array $forbidden
   *   The forbidden components.
   *
   * @return array
   *   The list of allowed values.
   */
  public static function allAllowedComponentValuesSelect(array $allowed = [], array $forbidden = []): array {
    $components = \Drupal::service(ComponentDiscovery::class)->findAllOrganisms();
    $options = [];
    foreach ($components as $component) {
      assert($component instanceof Component);
      $metadata = $component->getMetadata();
      $machine_name = $metadata->getMachineName();
      // If there are allowed components and this is not one of them, continue.
      if (!empty($allowed) && !in_array($machine_name, $allowed, TRUE)) {
        continue;
      }
      // If this is a forbidden component, continue.
      if (in_array($machine_name, $forbidden, TRUE)) {
        continue;
      }
      if (!isset($options[$metadata->getGroup()])) {
        $options[$metadata->getGroup()] = [];
      }
      $name = $component->getMetadata()->getName();
      $options[$metadata->getGroup()][$machine_name] = $name;
    }
    ksort($options);
    foreach ($options as $group => $opts) {
      natsort($options[$group]);
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function isEmpty() {
    return $this->component === NULL
      && $this->variant === NULL;
  }

  /**
   * {@inheritdoc}
   */
  public static function propertyDefinitions(FieldStorageDefinitionInterface $field_definition) {
    $properties['component'] = DataDefinition::create('string')
      ->setLabel(t('Component'));
    $properties['variant'] = DataDefinition::create('string')
      ->setLabel(t('Variant'));

    return $properties;
  }

  /**
   * {@inheritdoc}
   */
  public function getConstraints() {
    $constraints = parent::getConstraints();

    $options['component']['AllowedValues'] = array_keys(static::allAllowedComponentValues());
    $options['component']['NotBlank'] = [];

    $constraint_manager = \Drupal::typedDataManager()
      ->getValidationConstraintManager();
    $constraints[] = $constraint_manager->create('ComplexData', $options);
    return $constraints;
  }

  /**
   * {@inheritdoc}
   */
  public static function schema(FieldStorageDefinitionInterface $field_definition) {
    return [
      'columns' => [
        'component' => [
          'type' => 'varchar',
          'length' => 255,
        ],
        'variant' => [
          'type' => 'varchar',
          'length' => 255,
        ],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public static function generateSampleValue(FieldDefinitionInterface $field_definition) {
    return [
      'component' => array_rand(static::allAllowedComponentValues()),
      'variant' => 'my-variant',
    ];
  }

  /**
   * Returns allowed values for 'component' sub-field.
   *
   * @return array
   *   The list of allowed values.
   */
  public function allowedComponentValues(): array {
    return array_reduce(
      $this->allowedComponentValuesSelect(),
      static function (array $options, array $group) {
        return array_merge($options, $group);
      },
      []
    );
  }

  /**
   * Returns allowed values for 'component' sub-field when using a select.
   *
   * @return array
   *   The list of allowed values.
   */
  public function allowedComponentValuesSelect(): array {
    $settings = $this->getSettings();
    $allowed = array_keys(array_filter($settings['allowed'] ?? []));
    $forbidden = array_keys(array_filter($settings['forbidden'] ?? []));
    return static::allAllowedComponentValuesSelect($allowed, $forbidden);
  }

  /**
   * Allowed variants.
   *
   * @param $component_id string
   *   The component ID.
   *
   * @return string[]
   *   The variants.
   */
  public function allowedVariantValues(string $component_id): array {
    try {
      $component = \Drupal::service(ComponentDiscovery::class)->find($component_id);
      $variants = $component->getVariants();
      return array_combine($variants, $variants);
    }
    catch (ComponentNotFoundException $e) {
      return [];
    }
  }

}
