<?php

namespace Drupal\cl_selector_field\Plugin\Field\FieldWidget;

use Drupal\cl_components\Component\ComponentDiscovery;
use Drupal\cl_components\Exception\ComponentNotFoundException;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetBase;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;

/**
 * Defines the 'cl_selector_field_style_selector' field widget.
 *
 * @FieldWidget(
 *   id = "cl_selector_field_style_selector",
 *   label = @Translation("Style Selector"),
 *   field_types = {"cl_selector_field_style_selector"},
 * )
 */
class StyleSelectorWidget extends WidgetBase {

  /**
   * The discovery.
   *
   * @var \Drupal\cl_components\Component\ComponentDiscovery
   */
  private ComponentDiscovery $componentDiscovery;

  /**
   * The path where Drupal is installed.
   *
   * @var string
   */
  private string $appRoot;

  /**
   * {@inheritdoc}
   */
  public function __construct($plugin_id, $plugin_definition, FieldDefinitionInterface $field_definition, array $settings, array $third_party_settings, ComponentDiscovery $component_discovery, string $app_root) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->componentDiscovery = $component_discovery;
    $this->appRoot = $app_root;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    $component_discovery = $container->get(ComponentDiscovery::class);
    assert($component_discovery instanceof ComponentDiscovery);
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'], $configuration['third_party_settings'],
      $component_discovery,
      $container->getParameter('app.root')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    $item = $items[$delta] ?? NULL;
    $all_parents = array_merge(
      $element['#field_parents'],
      [$items->getName(), $delta, 'component']
    );
    $user_input = $form_state->getUserInput();
    $selected_component = NestedArray::getValue($user_input, $all_parents);
    $selected_component = !is_null($selected_component)
      ? $selected_component
      : ($item->component ?? NULL);
    $element['component'] = [
      '#type' => 'select',
      '#title' => $this->t('Component'),
      '#options' => ['' => $this->t('- Select a value -')] + $item->allowedComponentValuesSelect(),
      '#default_value' => $selected_component,
      '#ajax' => [
        'wrapper' => 'style-selector-wrapper',
        'method' => 'replaceWith',
        'callback' => [$this, 'ajaxDataHandler'],
      ],
    ];

    if ($selected_component) {
      $features = array_keys(array_filter($items->getSetting('features')));
      $allowed_variants = $item->allowedVariantValues($selected_component);
      if (in_array('variant', $features, TRUE) && !empty($allowed_variants)) {
        $current_variant = $items[$delta]->variant;
        $element['variant'] = [
          '#type' => 'select',
          '#title' => $this->t('Variant'),
          '#options' => [NULL => $this->t('- Default -'), ...$allowed_variants],
          '#default_value' => isset($allowed_variants[$current_variant]) ? $current_variant : NULL,
        ];
      }
      $element['documentation'] = $this->buildComponentInfo($selected_component);
    }

    $element['#theme_wrappers'] = ['container', 'form_element'];
    $element['#attributes']['class'][] = 'container-inline';
    $element['#attributes']['id'][] = 'container-inline-style-selector';
    $element['#prefix'] = '<div id="style-selector-wrapper">';
    $element['#suffix'] = '</div>';
    $element['#attributes']['class'][] = 'cl-components-style-selector-elements';
    $element['#attached']['library'][] = 'cl_selector_field/cl_components_style_selector';
    $element['#description'] = $this->t('💡 Are you missing a component from the dropdown? Make sure the component type is set to "organism" in the metadata.');

    return $element;
  }

  /**
   * Builds the component information.
   *
   * @param string $machine_name
   *   The machine name of the component.
   */
  private function buildComponentInfo(string $machine_name): array {
    try {
      $component = $this->componentDiscovery->find($machine_name);
    }
    catch (ComponentNotFoundException $exception) {
      return [];
    }
    $metadata = $component->getMetadata();
    $filename = $metadata->getThumbnailPath();
    $documentation = $metadata->getDocumentation() ?: $this->t('<em>No documentation available. Add a "README.md" and a "thumbnail.png" to the component folder to render them here.</em>');
    $alt = $this->t('Thumbnail for %component', ['%component' => $metadata->getName()]);
    return [
      '#type' => 'details',
      '#attributes' => ['class' => ['component-info']],
      '#title' => $this->t(
        'About %component <span class="component-status @status"></span>',
        [
          '%component' => $metadata->getName(),
          '@status' => $metadata->getStatus(),
        ]
      ),
      'thumbnail' => file_exists($filename) ? [
        'img' => [
          '#theme' => 'image',
          '#uri' => $filename,
          '#alt' => $alt,
          '#title' => $alt,
          '#prefix' => '<figure>',
        ],
        'caption' => [
          '#markup' => $alt,
          '#prefix' => '<figcaption>',
          '#suffix' => '</figcaption></figure>',
        ],
        'ruler' => ['#markup' => '<hr />'],
      ] : [],
      'docs' => [
        '#type' => 'container',
        'documentation' => ['#markup' => $documentation],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $violation, array $form, FormStateInterface $form_state) {
    return isset($violation->arrayPropertyPath[0]) ? $element[$violation->arrayPropertyPath[0]] : $element;
  }

  /**
   * {@inheritdoc}
   */
  public function massageFormValues(array $values, array $form, FormStateInterface $form_state) {
    foreach ($values as $delta => $value) {
      $component = $value['component'] ?? '';
      if ($component === '') {
        $values[$delta]['component'] = NULL;
        $values[$delta]['variant'] = NULL;
      }
      $variant = $value['variant'] ?? '';
      if ($variant === '') {
        $values[$delta]['variant'] = NULL;
      }
    }
    return $values;
  }

  /**
   * Handles the AJAX data.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form interface.
   *
   * @return array|mixed|null
   *   The results.
   */
  public function ajaxDataHandler(array &$form, FormStateInterface $form_state) {
    // Pop the last element so we render the whole thing.
    $array_parents = array_slice(
      $form_state->getTriggeringElement()['#array_parents'],
      0,
      -1
    );
    return NestedArray::getValue($form, $array_parents);
  }

}
